%
%   =====================================================================
%   Copyright (C) 2018  Peer Neubert, peer.neubert@etit.tu-chemnitz.de
%   
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%   
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%   
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.
%   =====================================================================
%
% Software release for the paper:
% "A Sequence-Based Neuronal Model for Mobile Robot Localization"
% Peer Neubert, Subutai Ahmad, and Peter Protzel, Proc. of German 
% Conference on Artificial Intelligence, 2018 
%
% Simplified Higher Order Sequence Memory from HTM.
% No learning in spatial pooler. Starting with an empty set of minicolumns, 
% each time a feedforward pattern is seen, that is not similar to any 
% existing minicolumn, create a new one witch responds to thinsg similar to
% this pattern.
%
classdef SimpleHTM < handle

properties
    params
    % feedforward connections: FF(:,j) is the set of indices into the input
    % SDR of the j-th minicolumn. So the activation of all minicolumns can
    % be computed by sum(SDR(FF))
    FF
    P % P(i,j) is the predicted-flag for the i-th cell in the j-th minicolumn
    prevP
        
    nCols
    
    burstedCol  % flag if this minicolumn bursts
    prevBurstedCol
    
    winnerCells % idx of all winning cells (in matrix of same size as P)
    activeCells % idx of all active cells (in matrix of same size as P)
    
    prevWinnerCells
    prevActiveCells
    
    % datastructure that hold the connections for predictions. It is a
    % cell array of matrices.  predictionConnections{i,j} is a two row matrix
    % with each column [idx in P; permanence] <--- now: only first row yet
    predictionConnections 
    
    % store for each column the number of iterations since last burst
    timeSinceLastBurst
    
end
   
methods
    function obj = SimpleHTM(params)
        if nargin<1
            obj.params.nCellPerCol = 10;
            obj.params.minColumnActivity = 35; % min overlap with SDR of a column
            obj.params.nInConPerCol = 40;
            obj.params.nColsPerPattern = 10;
        else
            obj.params = params;
        end
        
        obj.P = []; 
        obj.prevP = [];
        
        obj.burstedCol = [];
        obj.prevBurstedCol = [];
        obj.timeSinceLastBurst = [];
        
        obj.predictionConnections = {};        
        
        obj.nCols = 0;
                
    end
    
    % reset some internal variables
    function prepareNewIteration(obj)
        fprintf('  Prepare new iteration\n');
        obj.prevWinnerCells = obj.winnerCells;
        obj.winnerCells = [];
        obj.prevActiveCells = obj.activeCells;
        obj.activeCells = [];
        
        obj.prevP = obj.P;
        obj.prevBurstedCol = obj.burstedCol;        
        obj.timeSinceLastBurst = obj.timeSinceLastBurst +1;
        if obj.nCols>0
            obj.P = zeros(obj.params.nCellPerCol, obj.nCols);
            obj.burstedCol = zeros(obj.nCols);
        end
    end
    
    % inputConnections ... idx in a potential input SDR ,stored in obj.FF
    % return index of the ne column
    function newColIdx = createNewColumn(obj, inputConnections)
        for i=1:obj.params.nColsPerPattern
            obj.FF(:,end+1) = datasample(inputConnections, obj.params.nInConPerCol)';        
            obj.P(:,end+1) = zeros(obj.params.nCellPerCol, 1);       
            obj.prevP(:,end+1) = zeros(obj.params.nCellPerCol, 1);
            obj.predictionConnections(:, end+1) = cell(obj.params.nCellPerCol, 1);
            obj.burstedCol(:,end+1) = 0;
            obj.prevBurstedCol(:,end+1) = 0;
            obj.timeSinceLastBurst(end+1) = 0;
        end
        obj.nCols = obj.nCols+obj.params.nColsPerPattern;         
        newColIdx = [obj.nCols-obj.params.nColsPerPattern+1 : obj.nCols];
        
        fprintf('  Created %d new column(s), nCols is %d\n', obj.params.nColsPerPattern, obj.nCols);
    end
    
    % - Compare inputSDR to all minicolumns to find active minicolumns
    % - Search for predicted cells in active minicolumns and activate their
    %   predictions
    function compute(obj, inputSDR)
        obj.prepareNewIteration();
        
        % How similar is the input SDR to the pattern of the minicolumns?
        columnActivity = obj.computeColumnActivations(inputSDR);
        
        % Is there an activity above threshold? If yes, activate the most
        % active column, otherwise create a new one and make this the active one.
        maxActivity = max(columnActivity);
        if maxActivity > obj.params.minColumnActivity
            activeCols = find(columnActivity>obj.params.minColumnActivity);
        else
            sdrNonZeroIdx = find(inputSDR);
            activeCols = obj.createNewColumn(sdrNonZeroIdx);
        end
        
        % for each active column:
        % - mark all predicted cells as winnerCells
        % - if there was no predicted cell, chose one and activate all
        % predictions
        % - activate predictions of winnerCells
        %
        obj.activeCells = [];
        obj.winnerCells = [];
        for i=activeCols
            predictedIdx = find(obj.prevP(:,i)>0); 
            
            if isempty(predictedIdx)
                % if there are no predicted: burst (predict from all cells
                % and choose one winner cell)                
                winnerCell = obj.burst(i);
                obj.winnerCells = [obj.winnerCells, ...
                    sub2ind(size(obj.P), winnerCell, i)];
                obj.activeCells = [obj.activeCells, ...
                    sub2ind(size(obj.P), 1:size(obj.P,1), repmat(i,1,size(obj.P,1)))];
                fprintf('  Burst: Column %d, winner cell is %d\n', i, winnerCell);
            elseif length(predictedIdx)==1
                % if there is only one predicted cell, make this the
                % winner cell
                winnerCell = predictedIdx;
                obj.activatePredictions(i, winnerCell);                
                obj.winnerCells = [obj.winnerCells, ...
                    sub2ind(size(obj.P), winnerCell, i)];              
                obj.activeCells = [obj.activeCells, ...
                    sub2ind(size(obj.P), winnerCell, i)];
                
                fprintf('  One cell was predicted: Column %d, winner cell is %d\n', i, winnerCell);
            else
                % if there are multiple predicted cells, make all winner
                % cells  
                fprintf('  Multiple cells were predicted(%d):\n', length(predictedIdx))
                for j=predictedIdx'                    
                    obj.activatePredictions(i,j);
                    obj.winnerCells = [obj.winnerCells, ...
                    sub2ind(size(obj.P), j, i)];  
                    obj.activeCells = [obj.activeCells, ...
                        sub2ind(size(obj.P), j, i)];
                    fprintf('          Column %d, winner cell is %d\n', i, j);
                end
                
            end
            
        end
        
        % learn predictions
        obj.learnPredictions();
        
        % also predict newly learned predictions
        for columnIdx=1:obj.nCols
            if obj.burstedCol(columnIdx) == 1        
                for i=1:size(obj.P,1)
                    obj.activatePredictions(columnIdx, i);
                end
            end
        end
                
    end

    % given the set of currently winning cells and the set of previously
    % winning cells, create prediction connection 
    function learnPredictions(obj)
        for prevIdx = obj.prevWinnerCells
            for curIdx = obj.winnerCells     
                [prevCellIdx, prevColIdx] = ind2sub(size(obj.P), prevIdx);
                [curCellIdx, curColIdx] = ind2sub(size(obj.P), curIdx);
                              
                % check whether this minicolumn was bursting
                existingPredConFlag = obj.checkExistingPredCon(prevColIdx, curIdx);
                if ~existingPredConFlag || rand(1)<=obj.params.probAdditionalCon
                    obj.predictionConnections{prevIdx}(:, end+1) = curIdx;                     
                    obj.predictionConnections{prevIdx} = unique(obj.predictionConnections{prevIdx});
                    fprintf('learned [col, cell]: [%d, %d] --> [%d, %d]\n', prevColIdx, prevCellIdx, curColIdx, curCellIdx)
                else
                    %fprintf('not learned sice there is alread a connection\n');
                end
                
                
            end
        end
                
    end
    
    % Check if there already is an predicting connection from the previous
    % column to this active cell. This is used during bursting to prevent
    % learning multiple connections from one column to a single cell. In
    % this case, the new connection should go to a new cell of the current
    % collumn, to indicate the different context.
    function existingPredConFlag = checkExistingPredCon(obj, prevColIdx, curCellIdx)
        existingPredConFlag = sum([obj.predictionConnections{:,prevColIdx}]==curCellIdx)>0;
    end
    
    % Activate the predictions of all cells and identify the cell with the
    % fewest forward predictions to be the winning cell. winnerCellIdx is the 
    % index of this cell in the minicolumn
    function winnerCellIdx = burst(obj, columnIdx)
        obj.burstedCol(columnIdx) = 1;
        obj.timeSinceLastBurst(columnIdx) = 0;
        
        for i=1:size(obj.P,1)
            obj.activatePredictions(columnIdx, i);
        end
        
        % winnerCell: one of the cells with fewest existing forward predictions     
        nForwardPredictionsPerCell = cellfun(@length, obj.predictionConnections(:,columnIdx));
        
        % (slightly) inhibit winning cells from the last iteration a little bit
        for i=1:length(obj.prevWinnerCells)
            [cellIdx, colIdx] = ind2sub(size(obj.P), obj.prevWinnerCells(i));
            if colIdx==columnIdx
                nForwardPredictionsPerCell(cellIdx) = nForwardPredictionsPerCell(cellIdx) + obj.params.nCellPerCol;
            end
        end
        
        % [~, winnerCellIdx] = min(nForwardPredictionsPerCell);      
        candidateIdx = find(nForwardPredictionsPerCell==min(nForwardPredictionsPerCell));
        winnerCellIdx = obj.resolveTie( candidateIdx );
    end
    
    % randomly select one element to break a tie
    function idx = resolveTie(obj, x)
        idx = datasample(x,1);
    end
    
    % Increase number of predictions for all cells that are predicted from this
    % cell
    function activatePredictions(obj, colIdx, cellIdx)
        predIdx = obj.predictionConnections{cellIdx, colIdx};
        obj.P(predIdx) = obj.P(predIdx) + 1;        
    end
        
    % evaluate all connections between the input space and columns
    function columnActivity = computeColumnActivations(obj, inputSDR)
        columnActivity = sum(inputSDR(obj.FF));
    end
 
    
end

end % class