%
%   =====================================================================
%   Copyright (C) 2018  Peer Neubert, peer.neubert@etit.tu-chemnitz.de
%   
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%   
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%   
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.
%   =====================================================================
%
% Software release for the paper:
% "A Sequence-Based Neuronal Model for Mobile Robot Localization"
% Peer Neubert, Subutai Ahmad, and Peter Protzel, Proc. of German 
% Conference on Artificial Intelligence, 2018 
%
% Load sample data, e.g. from SDR simulator and run SimpleHTM.
% The activation of cells in the HTM minicolumn is used as output SDR 
% Show comparison to direct pairwise comparison of input descriptors.
function demo_SimpleHTM(D)

   
    % load data, D: each row is an SDR
    if ~exist('D', 'var')

        % without noise
        loadname = 'sim_data/seq_multi_loop_noise05_al5';
    
        % various noise levels
%     loadname = 'sim_data/seq_multi_loop_noise01_al0';
%     loadname = 'sim_data/seq_multi_loop_noise0_al1';
%     loadname = 'sim_data/seq_multi_loop_noise05_al5';
    
        D = dlmread([loadname '.txt']);
        GT_data = dlmread([loadname '_gt.txt'],',', 1,0);
    end
    
    n = 200; % number of used images
    
    % simple HTM parameters
    params.maxPredDepth = 0;
    params.probAdditionalCon = 0.05;
    params.nCellPerCol = 32;
    params.nInConPerCol = round(nnz(D>0.5) / size(D,1));
    params.minColumnActivity = round(0.25*params.nInConPerCol); % min overlap with SDR of a column --> 0.9
    params.nColsPerPattern = 10; 
    
    % run simple HTM
    t = tic; 
    htm = SimpleHTM(params);
    outputSDR = {};
    for i=1:min(n,size(D,1))
        fprintf('-------- ITERATION %d ---------\n', i);
        % skip empty vectors
        if nnz(D(i,:))==0
            fprintf('empty vector, skip\n');
            continue;
        end
        htm.compute(D(i,:));
        outputSDR{end+1} = htm.winnerCells;    
    end
    toc(t)
    
    % create output SDR matrix from HTM winner cell output
    ll = cellfun(@length, outputSDR);
    M = zeros(length(outputSDR), max(ll));
    for i=1:length(outputSDR)
        M(i, outputSDR{i}) = 1;               
    end
    
    % get similarity matrices for winner cell output and direct SDR input
    S1 = evaluateSimilarity(M);
    
    % for comparison: pairwise similarity of input descriptors
    S2 = evaluateSimilarity(D(1:n,:));
    
    figure('name', 'rSHOSM'); imshow(S1,[]);
    figure('name', 'Pairwise'); imshow(S2,[]);
    
    % Create ground truth and show precision-recall curves
    GT = zeros(size(S1));
    for i=1:size(GT,1)
        for j=i:size(GT,2)
            GT(i,j) = sum(GT_data(i,:) ~= GT_data(j,:))==0;
        end
    end
    
    [P1, R1] = createPR(S1, GT);    
    [P2, R2] = createPR(S2, GT);
        
    % show results
    figure();
    h1 = plot(R1,P1, 'g-'); hold on;
    h2 = plot(R2,P2, 'r-');
    axis([0 1 0 1])
    xlabel('Recall')
    ylabel('Precision');
    legend([h1, h2], {'Winner cells', 'Input SDRs'}, 'location', 'southwest');
    
    % save precision-recall values
    %save('results/pr_multiloop_hosm.mat', 'P1', 'R1');
    %save('results/pr_multiloop_pairs.mat', 'P2', 'R2');
end

% Create a precision recall curve
function [P, R] = createPR(S, GT)
    
    R=[0];
    P=[1];

    for i=1:-0.01:0
        B = S>=i;
        
        TP = nnz( (GT==1).*(B==1) );
        FN = nnz( (GT==1).*(B==0) );
        FP = nnz( (GT==0).*(B==1) );
        
        P(end+1) = TP/(TP + FP);
        R(end+1) = TP/(TP + FN);
        
    end

    R(end+1) = 1;
    P(end+1) = 0;
        
end

% Evaluate pairwise overlap similarity of SDRs in D
% D ... measurement/encodeing data matrix, each row is a sample
function S = evaluateSimilarity(D)
    
    t = 0;  
    
    nonZeroEntries = cell(size(D,1),1);
    for i=1:size(D,1)
        nonZeroEntries{i} = find(D(i,:)>t);
    end

    % pairwise comparison
    S = zeros(size(D,1));
    for idx1 = 1:size(D,1)
        nonZeroEntries1 = nonZeroEntries{idx1};
        for idx2 = idx1:size(D,1)
            nonZeroEntries2 = nonZeroEntries{idx2};

            commonEntries = intersect(nonZeroEntries1,nonZeroEntries2);
            meanNEntries = mean([numel(nonZeroEntries1), numel(nonZeroEntries2)]);
            if meanNEntries ~= 0
                S(idx1, idx2) = numel(commonEntries) / meanNEntries;
            else
                S(idx1, idx2) = 0;
            end
        end
    end
end
