%
%   =====================================================================
%   Copyright (C) 2018  Peer Neubert, peer.neubert@etit.tu-chemnitz.de
%   
%   This program is free software: you can redistribute it and/or modify
%   it under the terms of the GNU General Public License as published by
%   the Free Software Foundation, either version 3 of the License, or
%   (at your option) any later version.
%   
%   This program is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%   
%   You should have received a copy of the GNU General Public License
%   along with this program.  If not, see <http://www.gnu.org/licenses/>.
%   =====================================================================
%
% Software release for the paper:
% "A Sequence-Based Neuronal Model for Mobile Robot Localization"
% Peer Neubert, Subutai Ahmad, and Peter Protzel, Proc. of German 
% Conference on Artificial Intelligence, 2018 
%
% Run a simulator to create an output sequence together with ground truth
function run_simulator
    
    % set parameters
    params.savename = 'sim_data/test'; % a .txt, _gt.txt and a .png are created 
    params.env.size = [100, 100]; % [20,20]; % size of environment: [h,w]
    params.pathprop.random = 0; % create random path or use hard coded path 
    params.pathprop.start = [50,50]; %[8 3]; % start position of the path (x,y)
    % motion sequence: 1 up, 2 left, 3 down , 4 right, 0 non
    params.pathprop.motion = [1 1 1 1 1 1 1 1 4 4 4 4 3 3 3 3 3 2 2 2 2 1 1 1 1 1 ...
                                              4 4 4 4 3 3 3 3 3 2 2 2 2 1 1 1 1 1 ...
                                              4 4 4 4 3 3 3 3 3 2 2 2 2 1 1 1 1 1 ...
                                              4 4 4 4 3 3 3 3 3 2 2 2 2 1 1 1 1 1 ...
                                              4 4 4 4 3 3 3 3 3 2 2 2 2 1 1 1 1 1 ...
                                              4 4 4 4 3 3 3 3 3 2 2 2 2 1 1 1 1 1 ...
                                              4 4 4 4 3 3 3 3 3 2 2 2 2 1 1 1 1 1 ...
                                              4 4 4 4 3 3 3 3 3 2 2 2 2 1 1 1 1 1 ...
                                              4 4 4 4 3 3 3 3 3 2 2 2 2 1 1 1 1 1 ...
                                              4 4 4 4 3 3 3 3 3 2 2 2 2 1 1 1 1 1 ...
                                              4 4 4 4 3 3 3 3 3 2 2 2 2 1 1 1 1 1 ...
                                              1 1 1 1 1];  
   
    params.sdr.length = 2048; % descriptor length
    params.sdr.nNonZero = 40; % expected number of non zero entries
    params.sdr.random = 1; % is the SDR assigned randomly or computed from the pose

    % amount of noise
    params.path.SDRNoiseRate = 0.50; % probability of moving a bit

    % amount of aliasing
    % each row: x1 y1 x2 y2 of places (x1,y1) and (x2,y2) that share an SDR
%     params.sdr.sameSDRPairs = [8,6,12,6] % one aliasing pair; 
%     params.sdr.sameSDRPairs = [8,6,12,6; ...  % 5 aliasing pairs
%                                8,7,12,7; ...
%                                8,8,12,8; ...
%                                8,9,12,9; ...
%                                8,10,12,10];                 
    params.sdr.sameSDRPairs = []; % no aliasing 
    
    
    % for repeatability
    rng(0);
    
    % create world
    SDR = assignSDRtoPlaces(params);
    
    % travel path
    P = travelPath(params, SDR);
    
    % visualize 
    visualizePath(params, P);
    
    % save
    dlmwrite([params.savename '_gt.txt'], '# each row: x y','delimiter','');
    dlmwrite([params.savename '_gt.txt'],  P(: ,1:2), '-append');
    dlmwrite([params.savename '.txt'],  P(: ,3:end));
    
end


% show path on figure and save to disk using params.savename
function visualizePath(params,P)

    figure();
    % draw world
    [XX, YY] = meshgrid(1:params.env.size(2), 1:params.env.size(1));
    plot(XX,YY,'+k');
    hold on;
    
    % draw path
    x = P(:,1)+0.075*randn(size(P,1),1);
    y = P(:,2)+0.075*randn(size(P,1),1);
    z = zeros(size(x));
    col = 1:length(x);  % This is the color, vary with x in this case.
    surface([x';x'],[y';y'],[z';z'],[col;col],...        
        'facecol','no',...
        'edgecol','interp',...
        'linew',2);
    
    plot(P(:,1), P(:,2), '*')
    plot(P(1,1), P(1,2), 'o')
    axis equal;
    
    print(params.savename, '-dpng');
end

% P ... each row is [# x y SDR]
function P = travelPath(params,SDR)
    
    if params.pathprop.random
        error('Not yet implemented');
    else
        P = zeros(length(params.pathprop.motion), 2+params.sdr.length);
        curPose = params.pathprop.start;
        P(1,:) = [curPose reshape(SDR(curPose(2), curPose(1), :), 1, params.sdr.length)];
        for i=1:length(params.pathprop.motion)
            % apply motion and append data to P
            curPose = move(curPose, params.env.size, params.pathprop.motion(i));
            
            curSDR = reshape(SDR(curPose(2), curPose(1), :), 1, params.sdr.length);
            if params.path.lightNoiseFlag
                noisyCurSDR = applyNoiseOnSDRLight(params, curSDR);
            else
                noisyCurSDR = applyNoiseOnSDR(params, curSDR);
            end
            P(i+1,:) = [curPose noisyCurSDR];
        end
    end
end

function SDR = applyNoiseOnSDR(params, SDR)
    idx = find(SDR);
    for i=1:length(idx)
        switchFlag = rand(1) <= params.path.SDRNoiseRate;
        if switchFlag
            SDR(idx(i)) = 0;
            SDR(randi(length(SDR),1,1)) = 1;
        end
    end
end

% Apply motion command: 1 up, 2 left, 3 down , 4 right, 0 non
% Borders beyond environment borders have no effect
function newPose = move(oldPose, envSize, motionCommand) 
    newPose = oldPose + (motionCommand==1)*[ 0  1] ... 
                      + (motionCommand==2)*[-1  0] ... 
                      + (motionCommand==3)*[ 0 -1] ...
                      + (motionCommand==4)*[ 1  0];
    newPose(1) = min(max(1, newPose(1)),envSize(2));
    newPose(2) = min(max(1, newPose(2)),envSize(1));                  
end

% Assign an SDR  to each place
% SDR ... [nx,ny,lengthSDR] binary matrix; nx,ny is size of world
function SDR = assignSDRtoPlaces(params)
    if params.sdr.random
        % create a set of random SDRs
        SDR = rand(params.env.size(1), ...
            params.env.size(2), params.sdr.length);
        SDR = SDR < params.sdr.nNonZero/params.sdr.length;
        
        % enforce some places to share the same SDR
        for i=1:size(params.sdr.sameSDRPairs,1)
            x1 = params.sdr.sameSDRPairs(i,1);
            y1 = params.sdr.sameSDRPairs(i,2);
            x2 = params.sdr.sameSDRPairs(i,3);
            y2 = params.sdr.sameSDRPairs(i,4);
            SDR(y2,x2,:) = SDR(y1,x1,:);
        end
        
    else
        error('Not yet implemented');
    end
    
end


